
import java.util.Vector;

public class GEO implements Consts
{
   public int MinWeight;
   public int MaxWeight;
   public int EstPackagesHB;
   public int nEstPackagesHB;
   public int BaseCountStart;
   public int PackagesDelivered;

   public Yngwie yngwie;
   public int Width;
   public int Height;
   public Cell[][] Grid;
   public Vector Bases;
   public Vector Players;
   public int MapModus;

   public GEO(Yngwie y)
   {
      yngwie = y;
      Bases = new Vector(20,50);
      Players = new Vector(20,50);
      MapModus = cNoneMapped;
      MinWeight = 1000000000;
      MaxWeight = 1;
      EstPackagesHB = cUnknown;
      nEstPackagesHB = 0;
      BaseCountStart = 1;
      PackagesDelivered = 0;
   }

   public int IndexOfPlayer(int ID)
   {
      Player p;
      for (int i = 0; i < Players.size();i++)
      {
         p = (Player) Players.elementAt(i);
         if (p.ID == ID)
            return i;
      }
      return -1;
   }

   public Player PLOccupied(int x, int y)
   {
      Player p = null;
      for (int i = 0; i < Players.size();i++)
      {
         p = (Player) Players.elementAt(i);
         if ((p.X == x) && (p.Y == y))
            return p;
      }
      return null;
   }

   public DPackage getPackage(int ID)
   {
      HomeBase hb;
      Player pl;
      int idx;
      for (int i = 0; i < Bases.size();i++)
      {
         hb = (HomeBase) Bases.elementAt(i);
         idx = hb.IndexOf(ID);
         if (idx != -1)
            return (DPackage) hb.Packages.elementAt(idx);
      }
      for (int i = 0; i < Players.size();i++)
      {
         pl = (Player) Players.elementAt(i);
         idx = pl.IndexOf(ID);
         if (idx != -1)
            return (DPackage) pl.Packages.elementAt(idx);
      }
      return null;
   }

   public void UpdatePlayers(String s)
   {
      int startpos = 1;
      int p;
      int ID;
      int idx;
      String t;
      Player player;
      boolean ready = s.length() == 0;

      for (int i = 0; i < Players.size();i++)
      {
         player = (Player) Players.elementAt(i);
         player.Updated = false;
      }

      while (!ready)
      {
         p = s.indexOf("#",startpos);
         if (p == -1)
         {
            t = s.substring(startpos,s.length());
            t.trim();
            ready = true;
         }
         else
         {
            t = s.substring(startpos,p);
            t.trim();
         }

         ID = My.readInteger(t,0);
         idx = IndexOfPlayer(ID);
         if (idx == -1)
         {
            player = new Player(ID);
            Players.add(player);
            if (ID == yngwie.ID)
               yngwie.self = player;
         }
         else
            player = (Player) Players.elementAt(idx);

         player.Update(t);
         startpos = p+1;
      }

      int i = 0;
      while (i < Players.size())
      {
         player = (Player) Players.elementAt(i);
         if (! player.Updated)
         {
            player.Dead = true;
            Players.removeElementAt(i);
         }
         else
            i++;
      }
   }


   public void UpdatePackage(int ID, int x, int y, int w, int h, int destX, int destY)
   {
      boolean found = false;
      DPackage p = getPackage(ID);
      HomeBase hb = HBoccupied(x,y);
      if ((w != cUnknown) && (w < MinWeight))
         MinWeight = w;
      if (w > MaxWeight)
         MaxWeight = w;
      if (p != null)
      {
         p.X = x;
         p.Y = y;
         if (p.DestX == cUnknown)
         {
            p.DestX = destX;
            p.DestY = destY;
         }
         if ((p.Weight == cUnknown) && (w != cUnknown))
            p.Weight = w;
         if (p.Hold != h) // hij is geswitched
         {
            Player pl;
            if (p.Hold > 0)
               pl = (Player) Players.elementAt(IndexOfPlayer(p.Hold));
            else
               pl = (Player) Players.elementAt(IndexOfPlayer(h));

            if (h == -1)
            {
               pl.Packages.removeElementAt(pl.IndexOf(ID));
               if ((p.DestX != cUnknown) && (p.DestX == x) && (p.DestY == y))
               {
                  PackagesDelivered++;
               }
               else
               {
                  if (p.DestX == cUnknown)
                     p.MightBeGone = true;

                  if (hb == null)
                  {
                     hb = new HomeBase(x,y,Grid[x][y]);
                     hb.Checked = true; // deze basissen bevatten geen verborgen packages
                     Bases.add(hb);
                     NewMap(hb.X,hb.Y);
                  }
                  hb.Packages.add(p);
               }
            }
            else
            {
               pl.Packages.addElement(p);
               hb.Packages.removeElementAt(hb.IndexOf(ID));
               if ((hb.Checked) && (hb.Packages.size() == 0))
                  Bases.removeElement(hb);
            }
            p.Hold = h;
         }
      }
      else
      {
         p = new DPackage(ID,x,y,w,h,destX,destY);

         if (h == -1)
         {
            if (hb == null)
            {
               hb = new HomeBase(x,y,Grid[x][y]);
               hb.Checked = true; // deze basissen bevatten geen verborgen packages
               Bases.add(hb);
               NewMap(hb.X,hb.Y);
            }
            hb.Packages.add(p);
            if (destX != cUnknown)
               NewMap(destX,destY);
         }
         else
         {
            Player pl = (Player) Players.elementAt(IndexOfPlayer(h));
            pl.Packages.add(p);
         }
      }
   }

   public HomeBase HBoccupied(int x,int y)
   {
      HomeBase hb;
      for (int i = 0; i < Bases.size();i++)
      {
         hb = (HomeBase) Bases.elementAt(i);
         if ((hb.X == x) && (hb.Y == y))
            return hb;
      }
      return null;
   }

   public void Initiate()
   {
      // get Grid dimensions
      yngwie.WaitForReply();
      Width = Integer.parseInt(yngwie.inString.substring(0,yngwie.inString.indexOf(" ")));
      Height = Integer.parseInt(yngwie.inString.substring(yngwie.inString.indexOf(" ")+1));

      Grid = new Cell[Width][Height];
      // get Grid
      byte Type;
      int PassableCount = 0;
      char c;
      for (int i = 0; i < Height;i++)
      {
         yngwie.WaitForReply();
         if (yngwie.inString.length() != Width)
            System.out.println("error receiving row!");
//         System.out.println((i+1)+"  |   "+yngwie.inString);
         for (int j = 0; j < Width;j++)
         {
            c = yngwie.inString.charAt(j);
            if (c == '.')
            {
         	PassableCount++;
		Type = cPlain;
            }
            else if (c == '~')
               Type = cWater;
            else if (c == '#')
               Type = cWall;
            else if (c == '@')
            {
               PassableCount++;
               Type = cHomeBase;
            }
            else Type = cPlain;

            Grid[j][i] = new Cell(Type);
            if (Type == cHomeBase)
            {
               HomeBase hb = new HomeBase(j,i,Grid[j][i]);
               BaseCountStart++;
               Bases.add(hb);
            }
         }
      }

      for (int i = 0; i < Width;i++)
         for (int j = 0; j < Height;j++)
            if (Grid[i][j].Type == cHomeBase)
               NewMap(i,j);
   }

   public void NewMap(int x, int y)
   {
      if (Grid[x][y].Map == null)
      {
         Grid[x][y].Map = new short[Width][Height];
         for (int i = 0; i < Width;i++)
            for (int j = 0; j < Height;j++)
               Grid[x][y].Map[i][j] = cUnknown;
         BuildMap(Grid[x][y],(short) 0,x,y);
      }
   }

   public int MoneyToKill(){
      if (Players.size() <= 1)
         return 0;
      else
         return Math.min(20,(yngwie.Money - MoneyNeeded()) / (2 * (Players.size()-1)));
   }

   public int MoneyToDefend(){
      if (Players.size() <= 1)
         return 0;
      else
         return Math.min(50,(yngwie.Money - MoneyNeeded()) / (2 * (Players.size()-1)));
   }


   public int MoneyNeeded(){
      if (nEstPackagesHB == 0)
      {
         return (Width + Height) * 10 * BaseCountStart;
      }
      else
      {
         int MeanWeight = Math.max(1,((MaxWeight + MinWeight) / 2));
         int amount = Math.max(1,((EstPackagesHB-PackagesDelivered) / Math.max(1,(yngwie.MaxCap / MeanWeight))));
         return ((Width + Height) * amount * BaseCountStart) / nEstPackagesHB;
      }
   }

   public void BuildMap(Cell c,short dist,int x,int y)
   {
      if ((c.Map[x][y] == cUnknown) || (dist < c.Map[x][y]))
      {
         c.Map[x][y] = dist;
         if (dist < 32767)
         {
            dist++;
            if ((x > 0) && Grid[x-1][y].Passable())
               BuildMap(c,dist,x-1,y);
            if ((x < Width-1) && Grid[x+1][y].Passable())
               BuildMap(c,dist,x+1,y);
            if ((y > 0) && Grid[x][y-1].Passable())
               BuildMap(c,dist,x,y-1);
            if ((y < Height-1) && Grid[x][y+1].Passable())
               BuildMap(c,dist,x,y+1);
         }
      }
   }

/*   public void RemoveBase(int x, int y)
   {
      HomeBase hb;
      int i = 0;
      while (i < Bases.size())
      {
         hb = (HomeBase) Bases.elementAt(i);
         if ((hb.X == x) &&  (hb.Y == y))
           Bases.removeElementAt(i);
         else
           i++;
      }
   }*/

   public HomeBase BestBase()
   {
      int bestidx = -1;
      short bestDist = Short.MAX_VALUE;
      short Dist;
      HomeBase hb;
      for (int i = 0;i<Bases.size();i++)
      {
         hb = (HomeBase) Bases.elementAt(i);
         Dist = Grid[hb.X][hb.Y].Map[yngwie.self.X][yngwie.self.Y];
         if ((Dist != cUnknown) && (Dist < bestDist))
         {
            bestDist = Dist;
            bestidx = i;
         }
      }
      if (bestidx > -1)
         return (HomeBase) Bases.elementAt(bestidx);
      else
         return null;
   }

   public int Manhattan(int x1, int y1, int x2, int y2)
   {
      if (!Grid[x1][y1].Passable() || !Grid[x1][y1].Passable())
         return cUnreachable;
      else
         return Math.abs(x1-x2) + Math.abs(y1-y2);
   }


   public int LethalCount(int X, int Y)
   {
      int score = 0;
      int e = Eval(X+1,Y);
      if (e == cLethal)
         score++;
      if (e == cEnemy)
         score+=2;
      e = Eval(X-1,Y);
      if (e == cLethal)
         score++;
      if (e == cEnemy)
         score+=2;
      e = Eval(X,Y-1);
      if (e == cLethal)
         score++;
      if (e == cEnemy)
         score+=2;
      e = Eval(X,Y+1);
      if (e == cLethal)
         score++;
      if (e == cEnemy)
         score+=2;
      return score;
   }

   public int Eval(int X, int Y)
   {
      if ((X < 0) || (X >= Width) || (Y < 0) || (Y >= Height) || (Grid[X][Y].Type == cWall))
         return cUnpassable;
      if (Grid[X][Y].Lethal())
         return cLethal;
      Player pl= PLOccupied(X,Y);
      if ((pl != null) && (pl.ID != yngwie.ID))
         return cEnemy;
      return cNothing;
   }


/*   public int FastRectDistance(int x1, int y1, int x2, int y2)
   {
      if (x1 > x2)
      {
         if (y1 > y2)
         {
            if (
      return -1;
   }*/
}
